/* init.c by Adam Rogoyski <apoc@laker.net> Temperanc on EFNet irc
 * Copyright (C) 1998, 1999 Adam Rogoyski
 * --- GNU General Public License Disclamer ---
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */


#include "hexedit.h"
#include <assert.h>

#if defined(HAVE_GETOPT_H) && defined(HAVE_GETOPT_LONG)
#include <getopt.h>
#else
#include "../gnu/getopt.h"
#include "../gnu/getopt.c"
#include "../gnu/getopt1.c"
#endif

#ifdef __linux__
#include <linux/hdreg.h>
#endif

extern char **environ;


void
init (int argc, char **argv)
{
   unsigned char strbuf[DEFAULT_READ_BUFFER];
   int i = 0;
   FILE *fp = NULL;
   int len = 0;
   int result = 0;
   unsigned long filled = 0;
   int next_read = 0;
   struct stat filestat;
   void *p;

   int option_index = 0;
   struct option long_options[] = 
   {
      { "help",     0, 0, 0 },
      { "highbit",  0, 0, 0 },
      { "alltext",  0, 0, 0 },
      { "readonly", 0, 0, 0 },
      { "version",  0, 0, 0 },
      { "disk",     0, 0, 0 },
      { "buffer",   0, 0, 0 },
      { "quiet",    0, 0, 0 },
      { "force",    0, 0, 0 },
      {  NULL,      0, 0, 0 }
   };

   Globals.modified   = 0;
   Globals.filesize   = 0;
   Globals.print_mode = REGULAR_PRINT;
   Globals.buffsize   = DEFAULT_BUFFER_SIZE;
   Globals.fixed_disk = 0;
   Globals.fullsize   = 0;
   Globals.buf_front  = 0;
   Globals.buf_end    = 0;
   Globals.beeping    = 1;

      /* process environemnt variable first, then command line options */

   i = 0;
   while (*environ)
   {
      if (!strncasecmp (*environ, "HEXEDIT", 7))
      {
         while (**environ)
         {
            switch (**environ)
            {
               case '8':
                  Globals.print_mode = HIGH_ASCII_PRINT;
                  break;

               case 'a':
               case 'A':
                  Globals.print_mode = ALL_PRINT;
                  break;

               case 'b':
               case 'B':
                  Globals.buffsize = 0;
                  Globals.fullsize = 1;
                  break;

               case 'f':
               case 'F':
                  Globals.modified = 0;
                  break;

               case 'r':
               case 'R':
                  Globals.modified = READ_ONLY;
                  break;

               case 'd':
               case 'D':
                  Globals.fixed_disk = 1;
                  Globals.modified = READ_ONLY; /* need force as well */
                  break;

               case 'q':
               case 'Q':
                  Globals.beeping = 0;
                  break;
            }
            ++*environ;
         }
         break;
      }
      environ++;
      i++;
   }
   environ -= i;

   while ((i = getopt_long (argc, argv, "8abdfhqrv", long_options,
              &option_index)) != EOF)
   {
      switch (i)
      {
         case 0:
            switch (option_index)
            {
               case 0:
                  usage (*argv);
                  exit (EXIT_SUCCESS);

               case 1:
                  Globals.print_mode = HIGH_ASCII_PRINT;
                  break;

               case 2:
                  Globals.print_mode = ALL_PRINT;
                  break;

               case 3:
                  Globals.modified = READ_ONLY;
                  break;

               case 4:
                  printf ("%s\n", VERSION);
                  exit (EXIT_SUCCESS);

               case 5:
                  Globals.fixed_disk = 1;
                  Globals.modified = READ_ONLY; /* need force as well */
                  break;

               case 6:
                  Globals.buffsize = 0;
                  Globals.fullsize = 1;
                  break;

               case 7:
                  Globals.beeping = 0;
                  break;

               case 8:
                  Globals.modified = 0;
                  break;
            }
            break;

         case '8':
            Globals.print_mode = HIGH_ASCII_PRINT;
            break;

         case 'a':
            Globals.print_mode = ALL_PRINT;
            break;

         case 'b':
            Globals.buffsize = 0;
            Globals.fullsize = 1;
            break;

         case 'd':
            Globals.fixed_disk = 1;
            Globals.modified = READ_ONLY; /* need force as well*/
            break;

         case 'f':
            Globals.modified = 0;
            break;

         case 'h':
            usage (*argv);
            exit (EXIT_SUCCESS);

         case 'q':
            Globals.beeping = 0;
            break;

         case 'r':
            Globals.modified = READ_ONLY;
            break;

         case 'v':
            printf ("%s\n", VERSION);
            exit (EXIT_SUCCESS);
      }
   }

   i = 0;
   p = initscr ();
   if (p == NULL)
   {
      fprintf (stderr, "Cannot initialize screen - curses!\n");
      exit (EXIT_FAILURE);
   }
   keypad (stdscr, TRUE);
   scrollok (stdscr, FALSE);
   cbreak ();
   noecho ();
   refresh ();

   Globals.wstatus = newwin (1, COLS, 0, 0);
   if (Globals.wstatus == NULL)
      die_horribly ("Cannot open status window - curses!", NULL);

   scrollok (Globals.wstatus, FALSE);
   wattrset (Globals.wstatus, A_REVERSE);

   Globals.wmain = newwin (MAIN_HEIGHT, COLS, MAIN_TOP_LINE, 0);
   if (Globals.wmain == NULL)
      die_horribly ("Cannot open main window - curses!", NULL);

   scrollok (Globals.wmain, FALSE);

   Globals.whelp = newwin (1, COLS, LINES - 1, 0);
   if (Globals.whelp == NULL)
      die_horribly ("Cannot open help window - curses!", NULL);

   scrollok (Globals.whelp, FALSE);

#ifdef __PDCURSES__
   leaveok(Globals.wmain, FALSE);  /* Stupid macro */
#else
   leaveok (Globals.wmain, FALSE);
#endif


#ifndef __PDCURSES__
   signal (SIGWINCH, handleSigwinch);
#endif

   if (optind < argc)
   {
      fp = fopen (*(argv + optind), "rb");
      if (!fp)
         die_horribly (NULL, *(argv + optind));

      *(argv + 1) = *(argv + optind);
      optind = argc;
   }


   while (!fp && optind <= argc)
   {
      *(argv + 1) = fileSelect ();

      fp = fopen (*(argv + 1), "rb");
      if (!fp)
      {
#define FILE_ERR_WIDTH  50
#define FILE_ERR_HEIGHT 6
         WINDOW *wpopup = popupWindow (FILE_ERR_WIDTH, FILE_ERR_HEIGHT);

         werase (wpopup);
         box (wpopup, 0, 0);
         wmove (wpopup, 1, (FILE_ERR_WIDTH / 2)
                 - (strlen ("Cannot Open File") / 2));
         wprintw (wpopup, "Cannot Open FILE");
         wmove (wpopup, FILE_ERR_HEIGHT - 3, (FILE_ERR_WIDTH / 2) -
                (strlen (strerror (errno)) + strlen ("Reason: ")) / 2);
         wprintw (wpopup, "Reason: %s", strerror (errno));
         wrefresh (wpopup);
         getch ();
      }
      else
         optind++;
   }

   Globals.filename = *(argv + 1);

   if (!fp)
      die_horribly (NULL, *(argv + 1));


   result = stat (*(argv + 1), &filestat);
   if (result == -1)
      die_horribly (NULL, *(argv + 1));

   if ((filestat.st_size == 0) && !Globals.fixed_disk)
      die_horribly ("File is Zero Bytes - Nothing to Edit!", NULL);


#ifdef __linux__
   else if (Globals.fixed_disk)
   {
      struct hd_geometry fixed_disk;

      result = ioctl (fileno (fp), HDIO_GETGEO, &fixed_disk);
      if (result == -1)
         die_horribly (NULL, *(argv + 1));

      Globals.filesize = fixed_disk.heads * fixed_disk.sectors
                       * fixed_disk.cylinders * 512;
      filestat.st_size = DEFAULT_BUFFER_SIZE;
   }
#endif

      /* it would be zero if the -b, --buffer flag was given */
   if (!Globals.buffsize)
      Globals.buffsize = filestat.st_size;
   else
      extrabuf = 0;  /* no extrabuf since delete and insert not allowed */

   filebuf = malloc (Globals.buffsize + 1 + extrabuf);

   if (!filebuf)
      die_horribly (NOT_ENOUGH_MEMORY, NULL);

   memset (filebuf, 0x00, Globals.buffsize + 1 + extrabuf);

   i = 0;
   filled = 0;

   next_read = (DEFAULT_READ_BUFFER > Globals.buffsize) ?
                    Globals.buffsize
                  : DEFAULT_READ_BUFFER;

/*
erase (); refresh (); endwin ();
printf ("filled=%u  next_read=%d  filesize=%lu\n",
   filled, next_read, filestat.st_size);
*/

   memset (strbuf, 0, DEFAULT_READ_BUFFER);
   while ((result = fread (strbuf, 1, next_read, fp)) > 0)
   {
      if (ferror (fp))
         die_horribly (NULL, (*(argv + 1)));

      memcpy (filebuf + i, strbuf, result);
      i += result;
      if (i > filestat.st_size)
         die_horribly ("file has changed, restart", NULL);

      filled += result;
      if (filled < Globals.buffsize)
      {
         next_read = ((filled + DEFAULT_READ_BUFFER) > Globals.buffsize) ?
                       Globals.buffsize - filled
                     : DEFAULT_READ_BUFFER;
      }

/*
printf ("filled=%u  next_read=%d  filesize=%lu\n",
   filled, next_read, filestat.st_size);
*/
      if ((next_read <= 0) || (filled >= Globals.buffsize))
         break;
      memset (strbuf, 0, DEFAULT_READ_BUFFER);
   }
/*
printf ("done");exit(1);
*/
/*
   *(filebuf + i) = '\0';
*/
/*
   result = fclose (fp);
   if (!fp)
   {
      perror ("fclose");
   }
*/

   if (!Globals.filesize)
      Globals.filesize = filestat.st_size - 1;
/*
erase (); refresh (); endwin ();
printf ("filled = %ld   filesize = %ld\n", filled -1, Globals.filesize);
exit (1);
assert ((filled -1) == (Globals.filesize));
*/
   initbuffer (fp, 0, filled - 1);
   Globals.mode = HEX_MODE;
   if (Globals.modified != READ_ONLY)
      Globals.modified = 0;
   Globals.tabb = 0;
   Globals.charset = ASCII_CHAR_SET;
   Globals.spacing = 1;
   len = strlen (*(argv + 1)) + 1;
   Globals.progname = malloc (len);
   strcpy (Globals.progname, *(argv + 1));
   signal (SIGINT, handleInterrupt);

   newlines = malloc (BOTTOM_LINE * sizeof (int));
   if (!newlines)
      die_horribly (NOT_ENOUGH_MEMORY, NULL);

      /* if the whole buffer is not read into memory, we need to
       * do hashing of all modifications to keep track of changes.
       * Also, for when not in fullsize mode, to keep bold changes.
       */
   init_hash ();

   UndoStack.s = 0;
   UndoStack.base = NULL;
   UndoStack.top = NULL;

   for (i = 0; i < COLS; i++)
      waddch (Globals.wstatus, ' ');
   printStatusWindow ();
   printHelpWindow ();

   drawdump (offset);
   wrefresh (Globals.wmain);
   cursor_y = MAIN_TOP_LINE;
   cursor_x = 10;
   move (cursor_y, cursor_x);
   refresh ();
}
